#
# Copyright (C) 2006-2007  Robey Pointer <robey@lag.net>
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
# 
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

import base64
import logging
import os
import struct
import sys
from logging import DEBUG, INFO, WARNING, ERROR, CRITICAL

from Crypto.Cipher import AES
from Crypto.Hash import SHA, SHA256, HMAC
from Crypto.PublicKey import DSA
from Crypto.Util.randpool import PersistentRandomPool, RandomPool
from otrpacket import OTRPacket, _inflate_long, _deflate_long


STATE_NO_SESSION = 0
STATE_STARTED_AKE = 1
STATE_COMPLETED_DH = 2
STATE_REVEALED_SIGNATURE = 3
STATE_DATA = 4

MSG_DH_COMMIT = 2
MSG_DATA = 3
MSG_DH_KEY = 10
MSG_REVEAL_SIGNATURE = 17
MSG_SIGNATURE = 18

# we only support otr v2 :)
SUPPORTED_VERSIONS = set([2])

MODULUS_STRING = """\
FFFFFFFF FFFFFFFF C90FDAA2 2168C234 C4C6628B 80DC1CD1
29024E08 8A67CC74 020BBEA6 3B139B22 514A0879 8E3404DD
EF9519B3 CD3A431B 302B0A6D F25F1437 4FE1356D 6D51C245
E485B576 625E7EC6 F44C42E9 A637ED6B 0BFF5CB6 F406B7ED
EE386BFB 5A899FA5 AE9F2411 7C4B1FE6 49286651 ECE45B3D
C2007CB8 A163BF05 98DA4836 1C55D39A 69163FA8 FD24CF5F
83655D23 DCA3AD96 1C62F356 208552BB 9ED52907 7096966D
670C354E 4ABC9804 F1746C08 CA237327 FFFFFFFF FFFFFFFF
"""
MODULUS = long(MODULUS_STRING.replace(' ', '').replace('\n', ''), 16)
GENERATOR = 2L


try:
    randpool = PersistentRandomPool(os.path.expanduser('~/.randpool'))
except:
    randpool = RandomPool()
randpool.randomize()


# FIXME temporary
def generate_pubkey():
    p = 1067993517960455041197510853084776057301352263943770102959004351808977783848450923363266649265913L
    q = 1224804481542489503237073782691335656202922468197L
    g = 683716890263519010837633943103568376662530557512293301413281382107392067145656233989822305721339L
    x = 345816517838335209401807862664759693719721112695L
    y = 748772335955891549232295758065275780692247810193128739373440908369442647725617639772684128797192L
    return DSA.construct((y, g, p, q, x))
MY_PUBKEY = generate_pubkey()


class _Counter (object):
    "object that can be used as a counter for MODE_CTR ciphers"
    
    _pad = '\000' * 16
    
    def __init__(self, initial=0):
        self._value = initial
        
    def __call__(self):
        out = _deflate_long(self._value, False)
        self._value += 1
        return (self._pad + out)[-16:]


class CipherState (object):
    """
    DH key halves, computed values, and state kept for a pair of key ids during an
    active session or initial AKE.
    """

    def __init__(self):
        self.x = None
        self.gx = None
        self.gy = None
        # data session
        self.sendkey = None
        self.recvkey = None
        self.sendmackey = None
        self.recvmackey = None
        self.counter = 0
        self.peer_counter = 0
        # AKE
        self.ssid = None
        self.c = None
        self.cp = None
        self.m1 = None
        self.m1p = None
        self.m2 = None
        self.m2p = None

    def compute_gx(self):
        if self.gx is None:
            self.gx = pow(GENERATOR, self.x, MODULUS)

    def compute_for_data(self):
        if self.sendkey is not None:
            # already done
            return
        self.compute_gx()
        s = pow(self.gy, self.x, MODULUS)        
        s_mpi = OTRPacket.make('mpi', s)
        if self.gx > self.gy:
            sendbyte, recvbyte = '\x01', '\x02'
        else:
            sendbyte, recvbyte = '\x02', '\x01'
        self.sendkey = SHA.new(sendbyte + s_mpi).digest()[:16]
        self.recvkey = SHA.new(recvbyte + s_mpi).digest()[:16]
        self.sendmackey = SHA.new(self.sendkey).digest()
        self.recvmackey = SHA.new(self.recvkey).digest()

    def compute_for_ake(self):
        if self.c is not None:
            # already done
            return
        self.compute_gx()
        s = pow(self.gy, self.x, MODULUS)
        s_mpi = OTRPacket.make('mpi', s)
        self.ssid = _inflate_long(SHA256.new('\000' + s_mpi).digest()[:8], True)
        hash = SHA256.new('\001' + s_mpi).digest()
        self.c = hash[:16]
        self.cp = hash[16:32]
        self.m1 = SHA256.new('\002' + s_mpi).digest()
        self.m2 = SHA256.new('\003' + s_mpi).digest()
        self.m1p = SHA256.new('\004' + s_mpi).digest()
        self.m2p = SHA256.new('\005' + s_mpi).digest()

    

class OTRHandler (object):
    def send(self, text):
        "send to remote user"
        pass
    
    def display(self, text):
        "display to local user"
        pass

    def encrypted(self, pubkey, ssid):
        "when encryption begins, this is called to provide the pubkey of the peer."
        pass
    
    def incoming(self, text, state):
        "an encrypted message has arrived"
        pass


class OTRSession (object):
    def __init__(self, logname='otr', handler=None):
        self._logger = logging.getLogger(logname)
        self._handler = handler
        self._pubkey = MY_PUBKEY
        self.clear()

    def start(self):
        self._handler.send('?OTRv2?')
        self._otr_version = 2
        self._state = STATE_NO_SESSION
    
    def stop(self):
        if self._state == STATE_DATA:
            self.send_raw('\x00' + OTRPacket.make('tlv', (1, '')))
        self.clear()
    
    def clear(self):
        self._state = STATE_NO_SESSION
        self._otr_version = 0
        self._ssid = None
        # temporary key during AKE:
        self._r = None
        self._gx_data = None
        self._gx_hash = None
        self._blob = None
        self._blob_hash = None
        # ongoing key exchange state
        # these are always [our latest, our latest - 1][peer latest, peer latest - 1]
        self._key_state = [ [ None, None ], [ None, None ] ]
        self._keyid = 0
        self._peer_keyid = 0
        self._mac_reveal = ''
        # fragment state:
        self._fragment_k = 0
        self._fragment_n = 0
        self._fragment_buffer = ''
    
    def is_active(self):
        return self._state == STATE_DATA

    def parse(self, text, state=None):
        """
        stuff.
        """
        if not '?OTR' in text:
            return
        n = text.find('?OTR?')
        if n >= 0:
            self.check_version(text[n:])
            return
        n = text.find('?OTRv')
        if n >= 0:
            self.check_version(text[n:])
            return
        n = text.find('?OTR Error:')
        if n >= 0:
            self.error(text[n:])
            return
        n = text.find('?OTR,')
        if n >= 0:
            self.get_fragment(text[n + 5:])
            return
        n = text.find('?OTR:')
        if n >= 0:
            # discard any fragments
            self._fragment_buffer = ''
            self._fragment_k = 0
            self._fragment_n = 0
            text = text[n + 5:]
            n = text.find('.')
            if n >= 0:
                text = text[:n]
            try:
                data = base64.decodestring(text)
            except:
                self._handler.display('Corrupted OTR data.')
                return
            self.process(data, state)

    def check_version(self, text):
        if self._state != STATE_NO_SESSION:
            self._handler.display('WARNING: OTR session reset')
            self._state = STATE_NO_SESSION
            self.log(WARNING, 'OTR session reset!')

        versions = set()
        if text[4] == '?':
            versions.add(1)
        for ch in text[5:]:
            if ch == '2':
                versions.add(2)
            # other unknown versions: ignore
        agreed = SUPPORTED_VERSIONS.intersection(versions)
        if len(agreed) == 0:
            self.log(ERROR, 'No common OTR versions supported.')
            return
        self._otr_version = max(agreed)
        self.send_dh_commit()
    
    def error(self, text):
        # ?OTR Error: message
        message = text[11:].strip()
        self.log(ERROR, 'OTR error message: %s' % (message,))
        self._handler.display('OTR ERROR: %s' % (message,))
    
    def get_fragment(self, text):
        segments = text.split(',')
        if len(segments) < 3:
            self.log(ERROR, 'Corrupted OTR fragment.')
            return
        k = int(segment[0])
        n = int(segment[1])
        if (k == 0) or (n == 0) or (k > n):
            self.log(ERROR, 'Corrupted OTR fragment. (%d,%d)' % (k, n))
            return
        if k == 1:
            if len(self._fragment_buffer) > 0:
                self.log(WARNING, 'Discarding previous unfinished fragments.')
            self._fragment_k = k
            self._fragment_n = n
            self._fragment_buffer = segment[2]
            return
        if (n == self._fragment_n) and (k == self._fragment_k + 1):
            self._fragment_buffer += segment[2]
            self._fragment_k += 1
            if k == n:
                self.parse(self._fragment_buffer)
                # FIXME clear fragments
            return
        self.log(WARNING, 'Unexpected fragment (discarding) (%d,%d).' % (k, n))
        return
        
    def process(self, data, state):
        if len(data) < 3:
            self.log(ERROR, 'Corrupted OTR packet.')
            return
        packet = OTRPacket(data)
        version, ptype = packet.unpack('short byte')
        if self._otr_version == 0:
            self.log(DEBUG, 'Discarding OTR message from an unknown (prior?) context.')
            self._handler.display('Undecipherable message, possibly from a prior encrypted chat.')
            return
        if version != self._otr_version:
            self.log(ERROR, 'Corrupted OTR packet (wrong version = %d vs %d)' % (version, self._otr_version))
            return
        if ptype == MSG_DH_COMMIT:
            self.process_dh_commit(packet)
        elif ptype == MSG_DATA:
            self.process_data(packet, state)
        elif ptype == MSG_DH_KEY:
            self.process_dh_key(packet)
        elif ptype == MSG_REVEAL_SIGNATURE:
            self.process_reveal_signature(packet)
        elif ptype == MSG_SIGNATURE:
            self.process_signature(packet)

    def send_dh_commit(self):
        self._r = randpool.get_bytes(16)
        self.generate_starting_keys()
        gxmpi = OTRPacket.make('mpi', self._key_state[1][0].gx)
        
        # we only remember gx_hash in case our DH_COMMIT messages cross in the mail
        self._gx_data = self.encrypt_aes_ctr(self._r, 0, gxmpi)
        self._gx_hash = SHA256.new(gxmpi).digest()
        
        out = OTRPacket.make('short byte blob blob', self._otr_version, MSG_DH_COMMIT, self._gx_data, self._gx_hash)
        self._send(out)
        self._state = STATE_STARTED_AKE
    
    def send_dh_key(self):
        out = OTRPacket.make('short byte mpi', self._otr_version, MSG_DH_KEY, self._key_state[1][0].gx)
        self._send(out)
        self._state = STATE_COMPLETED_DH

    def send_reveal_signature(self):
        self._key_state[1][0].compute_for_ake()
        data = OTRPacket.make('mpi mpi pubkey int', self._key_state[1][0].gx, self._key_state[1][0].gy, self._pubkey, self._keyid - 1)
        mb = HMAC.HMAC(self._key_state[1][0].m1, data, SHA256).digest()
        xb = OTRPacket.make('pubkey int sig', self._pubkey, self._keyid - 1, self.sign(mb))
        xbe = self.encrypt_aes_ctr(self._key_state[1][0].c, 0, xb)
        mac = HMAC.HMAC(self._key_state[1][0].m2, OTRPacket.make('blob', xbe), SHA256).digest()
        out = OTRPacket.make('short byte blob blob mac', self._otr_version, MSG_REVEAL_SIGNATURE, self._r, xbe, mac)
        self._send(out)
        self._state = STATE_REVEALED_SIGNATURE
    
    def send_signature(self):
        data = OTRPacket.make('mpi mpi pubkey int', self._key_state[1][0].gx, self._key_state[1][0].gy, self._pubkey, self._keyid - 1)
        ma = HMAC.HMAC(self._key_state[1][0].m1p, data, SHA256).digest()
        xa = OTRPacket.make('pubkey int sig', self._pubkey, self._keyid - 1, self.sign(ma))
        xae = self.encrypt_aes_ctr(self._key_state[1][0].cp, 0, xa)
        mac = HMAC.HMAC(self._key_state[1][0].m2p, OTRPacket.make('blob', xae), SHA256).digest()[:20]
        out = OTRPacket.make('short byte blob mac', self._otr_version, MSG_SIGNATURE, xae, mac)
        self._send(out)
        self._state = STATE_DATA
        self._ssid = self._key_state[1][0].ssid
        self._handler.encrypted(self._peer_pubkey, self._ssid)
    
    def process_dh_commit(self, packet):
        data, hash = packet.unpack('blob blob')
        if self._state == STATE_STARTED_AKE:
            # already started AKE; we're in a race.  largest hashed gx wins.
            my_gxh = _inflate_long(self._gx_hash, True)
            their_gxh = _inflate_long(hash, True)
            if my_gxh > their_gxh:
                # ignore and re-transmit DH_COMMIT
                out = OTRPacket.make('short byte blob blob', self._otr_version, MSG_DH_COMMIT, self._gx_data, self._gx_hash)
                self._send(out)
                return
            # forget old DH_COMMIT and treat this as a new beginning
            self.clear()
        elif self._state == STATE_COMPLETED_DH:
            # use new values, but send old DH_KEY
            self._blob = data
            self._blob_hash = hash
            self.send_dh_key()
            return
        elif self._state == STATE_REVEALED_SIGNATURE:
            # something got confused; clear out and back up.
            self.clear()

        if self._state != STATE_NO_SESSION:
            self.log(ERROR, 'Unexpected packet DH_COMMIT.')
            return
        self._blob = data
        self._blob_hash = hash
        self.generate_starting_keys()
        # now send our key
        self.send_dh_key()

    def process_dh_key(self, packet):
        if self._state not in (STATE_STARTED_AKE, STATE_REVEALED_SIGNATURE):
            self.log(DEBUG, 'Unexpected DH_KEY; ignoring.')
            return
        gy, = packet.unpack('mpi')
        if (gy < 2) or (gy > MODULUS - 2):
            self.log(ERROR, 'Peer\'s gy is out of range.')
            self.clear()
            return
        if self._state == STATE_REVEALED_SIGNATURE:
            # might be redundant; that's ok.
            if gy != self._key_state[0][0].gy:
                # ignore
                self.log(DEBUG, 'Redundant and different DH_KEY; ignoring.')
                return
            # re-transmit signature...
        self._key_state[0][0].gy = gy
        self._key_state[1][0].gy = gy
        self.send_reveal_signature()
    
    def process_reveal_signature(self, packet):
        if self._state != STATE_COMPLETED_DH:
            self.log(ERROR, 'Unexpected packet REVEAL_SIGNATURE.')
            return
        self._r, data, mac = packet.unpack('blob blob mac')
        # FIXME: why isn't mac ever used?
        gxmpi = self.decrypt_aes_ctr(self._r, 0, self._blob)
        gy, = OTRPacket.parse('mpi', gxmpi)
        check_hash = SHA256.new(OTRPacket.make('mpi', gy)).digest()
        if check_hash != self._blob_hash:
            self.log(ERROR, 'Corrupted hash on DH_COMMIT.')
            return
        self._key_state[0][0].gy = gy
        self._key_state[1][0].gy = gy
        self._key_state[1][0].compute_for_ake()
        xb = self.decrypt_aes_ctr(self._key_state[1][0].c, 0, data)
        self._peer_pubkey, self._peer_keyid, sig = OTRPacket.parse('pubkey int sig', xb)
        
        data = OTRPacket.make('mpi mpi pubkey int', self._key_state[1][0].gy, self._key_state[1][0].gx, self._peer_pubkey, self._peer_keyid)
        mac = HMAC.HMAC(self._key_state[1][0].m1, data, SHA256).digest()
        if not self._peer_pubkey.verify(_inflate_long(mac, True), sig):
            self.log(ERROR, 'Signature verification failed.')
            return
        
        # send my signature now
        self.send_signature()
    
    def process_signature(self, packet):
        if self._state != STATE_REVEALED_SIGNATURE:
            self.log(ERROR, 'Unexpected packet SIGNATURE.')
            return
        xae, mac = packet.unpack('blob mac')
        check_hash = HMAC.HMAC(self._key_state[1][0].m2p, OTRPacket.make('blob', xae), SHA256).digest()[:20]
        if check_hash != mac:
            self.log(ERROR, 'Corrupted hash on SIGNATURE.')
            return
        xa = self.encrypt_aes_ctr(self._key_state[1][0].cp, 0, xae)
        self._peer_pubkey, self._peer_keyid, sig = OTRPacket.parse('pubkey int sig', xa)
        data = OTRPacket.make('mpi mpi pubkey int', self._key_state[1][0].gy, self._key_state[1][0].gx, self._peer_pubkey, self._peer_keyid)
        ma = HMAC.HMAC(self._key_state[1][0].m1p, data, SHA256).digest()
        if not self._peer_pubkey.verify(_inflate_long(ma, True), sig):
            self.log(ERROR, 'Signature verification failed.')
            return
        self._state = STATE_DATA
        self._ssid = self._key_state[1][0].ssid
        self._handler.encrypted(self._peer_pubkey, self._ssid)

    def process_data(self, packet, state):
        flags, sender_keyid, recipient_keyid, dhy, counter, encdata, mac, old_keys = \
            packet.unpack('byte int int mpi ctr blob mac blob')
        key_index = self._keyid - recipient_keyid
        peer_key_index = self._peer_keyid - sender_keyid
        if (key_index not in (0, 1)) or (peer_key_index not in (0, 1)) or \
            ((peer_key_index == 1) and (self._key_state[0][1].gy is None)):
            # using a key we don't have
            self.log(ERROR, 'Illegal key id sender=%d, recipient=%d (current sender=%d, recipient=%d)' %
                     (sender_keyid, recipient_keyid, self._peer_keyid, self._keyid))
            self._handler.display('Illegal message encryption.')
            return
        ks = self._key_state[key_index][peer_key_index]
        ks.compute_for_data()
        data = OTRPacket.make('short byte byte int int mpi ctr blob',
                              self._otr_version, MSG_DATA, flags, sender_keyid, recipient_keyid, dhy, counter, encdata)
        my_mac = HMAC.HMAC(ks.recvmackey, data, SHA).digest()
        if my_mac != mac:
            self.log(ERROR, 'Corrupted MAC on incoming data. (%r != %r)' % (mac, my_mac))
            self._handler.display('Corrupted message.')
            return
        counter = long(counter)
        if counter <= ks.peer_counter:
            self.log(ERROR, 'Counter decreasing on incoming data.')
            self._handler.display('Corrupted message.')
            return
        ks.peer_counter = counter
        data = self.decrypt_aes_ctr(ks.recvkey, counter << 64, encdata)
        
        if key_index == 0:
            # they're using our latest key half!  make a new one...
            if self._key_state[1][0].sendmackey is not None:
                self._mac_reveal += self._key_state[1][0].sendmackey
            if self._key_state[1][0].recvmackey is not None:
                self._mac_reveal += self._key_state[1][0].recvmackey
            if self._key_state[1][1].sendmackey is not None:
                self._mac_reveal += self._key_state[1][1].sendmackey
            if self._key_state[1][1].recvmackey is not None:
                self._mac_reveal += self._key_state[1][1].recvmackey
            self._key_state[1] = self._key_state[0]
            self._key_state[0] = [ CipherState(), CipherState() ]
            x = self.generate_dh()
            self._key_state[0][0].x = x
            self._key_state[0][1].x = x
            self._key_state[0][0].gy = self._key_state[1][0].gy
            self._key_state[0][1].gy = self._key_state[1][1].gy
            self._keyid += 1
        
        if peer_key_index == 0:
            # we got a new key half!
            if self._key_state[0][1].sendmackey is not None:
                self._mac_reveal += self._key_state[0][1].sendmackey
            if self._key_state[0][1].recvmackey is not None:
                self._mac_reveal += self._key_state[0][1].recvmackey
            if self._key_state[1][1].sendmackey is not None:
                self._mac_reveal += self._key_state[1][1].sendmackey
            if self._key_state[1][1].recvmackey is not None:
                self._mac_reveal += self._key_state[1][1].recvmackey
            self._key_state[0][1] = self._key_state[0][0]
            self._key_state[1][1] = self._key_state[1][0]
            self._key_state[0][0] = CipherState()
            self._key_state[1][0] = CipherState()
            self._key_state[0][0].x = self._key_state[0][1].x
            self._key_state[1][0].x = self._key_state[1][1].x
            self._key_state[0][0].gy = dhy
            self._key_state[1][0].gy = dhy
            self._peer_keyid += 1
        
        if '\x00' in data:
            # TLVs!
            data, tlv = data.split('\x00', 1)
            p = OTRPacket(tlv)
            while p.remaining() > 0:
                (t, val), = p.unpack('tlv')
                if t == 0:
                    # padding
                    continue
                elif t == 1:
                    # disconnect
                    self.log(INFO, 'Disconnected from OTR.')
                    self.clear()

        # might have just been TLVs
        if len(data) > 0:
            self._handler.incoming(data, state)

    def send(self, text):
        # no TLVs attached
        self.send_raw(text + '\x00')

    def send_raw(self, text):
        sender_keyid = self._keyid - 1
        recipient_keyid = self._peer_keyid
        ks = self._key_state[1][0]
        ks.compute_for_data()
        
        if ks.counter == 0:
            counter = _inflate_long(randpool.get_bytes(7), True)
        else:
            counter = ks.counter
        ks.counter = counter + 1
        encdata = self.encrypt_aes_ctr(ks.sendkey, counter << 64, text)
        self._key_state[0][0].compute_gx()
        data = OTRPacket.make('short byte byte int int mpi ctr blob',
                              self._otr_version, MSG_DATA, 0, sender_keyid, recipient_keyid,
                              self._key_state[0][0].gx, counter, encdata)
        mac = HMAC.HMAC(ks.sendmackey, data, SHA).digest()
        data += OTRPacket.make('mac blob', mac, self._mac_reveal)
        self._mac_reveal = ''
        self._send(data)

    @staticmethod
    def generate_dh():
        return _inflate_long(randpool.get_bytes(40), True)

    def generate_starting_keys(self):
        self._keyid = 2
        x1, x2 = self.generate_dh(), self.generate_dh()
        self._key_state = [ [ CipherState(), CipherState() ], [ CipherState(), CipherState() ] ]
        self._key_state[0][0].x = x1
        self._key_state[0][1].x = x1
        self._key_state[1][0].x = x2
        self._key_state[1][1].x = x2
        for i in range(2):
            for j in range(2):
                self._key_state[i][j].compute_gx()
    
    def dump_keys(self):
        # for debugging -- not currently used
        self.log(DEBUG, 'keyid=%d, peer_keyid=%d' % (self._keyid, self._peer_keyid))
        for i in range(2):
            for j in range(2):
                ks = self._key_state[i][j]
                self.log(DEBUG, 'keyset[%d][%d] = { x=%r, gx=%r, gy=%r }' % (i, j, ks.x, ks.gx, ks.gy))
        return

    @staticmethod
    def encrypt_aes_ctr(key, iv, data):
        return AES.new(key, AES.MODE_CTR, counter=_Counter(iv)).encrypt(OTRSession.pad(data, 16))[:len(data)]

    @staticmethod
    def decrypt_aes_ctr(key, iv, data):
        return AES.new(key, AES.MODE_CTR, counter=_Counter(iv)).decrypt(OTRSession.pad(data, 16))[:len(data)]

    def sign(self, data):
        # already hashed
        qsize = len(_deflate_long(self._pubkey.q, False))
        while True:
            k = _inflate_long(randpool.get_bytes(qsize), True)
            if (k > 2) and (k < self._pubkey.q):
                break
        r, s = self._pubkey.sign(_inflate_long(data, True), k)
        return r, s
        
    def _send(self, data):
        text = base64.encodestring(data).replace('\n', '')
        self._handler.send('?OTR:%s.' % (text,))

    def log(self, level, text):
        self._logger.log(level, text)

    @staticmethod
    def pad(data, n):
        extra = len(data) % n
        if extra == 0:
            return data
        return data + ('\000' * (n - extra))
