#
# Copyright (C) 2006-2007  Robey Pointer <robey@lag.net>
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
# 
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

import struct
from Crypto.PublicKey import DSA


PUBKEY_TYPE_DSA = 0


def _inflate_long(s, always_positive=False):
    "turns a normalized byte string into a long-int (adapted from Crypto.Util.number)"
    out = 0L
    negative = 0
    if not always_positive and (len(s) > 0) and (ord(s[0]) >= 0x80):
        negative = 1
    if len(s) % 4:
        filler = '\x00'
        if negative:
            filler = '\xff'
        s = filler * (4 - len(s) % 4) + s
    for i in range(0, len(s), 4):
        out = (out << 32) + struct.unpack('>I', s[i:i+4])[0]
    if negative:
        out -= (1L << (8 * len(s)))
    return out


def _deflate_long(n, add_sign_padding=True):
    "turns a long-int into a normalized byte string (adapted from Crypto.Util.number)"
    # after much testing, this algorithm was deemed to be the fastest
    s = ''
    n = long(n)
    while (n != 0) and (n != -1):
        s = struct.pack('>I', n & 0xffffffffL) + s
        n = n >> 32
    # strip off leading zeros, FFs
    for i in enumerate(s):
        if (n == 0) and (i[1] != '\000'):
            break
        if (n == -1) and (i[1] != '\xff'):
            break
    else:
        # degenerate case, n was either 0 or -1
        i = (0,)
        if n == 0:
            s = '\000'
        else:
            s = '\xff'
    s = s[i[0]:]
    if add_sign_padding:
        if (n == 0) and (ord(s[0]) >= 0x80):
            s = '\x00' + s
        if (n == -1) and (ord(s[0]) < 0x80):
            s = '\xff' + s
    return s


def encode_pubkey(pubkey):
    return OTRPacket.make('short mpi mpi mpi mpi', PUBKEY_TYPE_DSA, pubkey.p, pubkey.q, pubkey.g, pubkey.y)

def decode_pubkey(data):
    packet = OTRPacket(data)
    keytype, p, q, g, y = packet.unpack('short mpi mpi mpi mpi')
    assert keytype == PUBKEY_TYPE_DSA
    key = DSA.construct((y, g, p, q))
    return key, packet._n


class OTRPacket (object):
    def __init__(self, data=''):
        self._data = data
        self._n = 0
    
    def __len__(self):
        return len(self._data)
    
    def __str__(self):
        return self._data

    def remaining(self):
        return len(self._data) - self._n

    def add(self, format, *args):
        args = list(args)
        for item in format.split(' '):
            if item == 'byte':
                self._data += struct.pack('>B', args.pop(0))
            elif item == 'short':
                self._data += struct.pack('>H', args.pop(0))
            elif item == 'int':
                self._data += struct.pack('>I', args.pop(0))
            elif item == 'ctr':
                self._data += struct.pack('>Q', args.pop(0))
            elif item == 'mpi':
                val = args.pop(0)
                mpi = _deflate_long(val, False)
                self._data += struct.pack('>i', len(mpi)) + mpi
            elif item == 'blob':
                val = args.pop(0)
                self._data += struct.pack('>i', len(val)) + val
            elif item == 'pubkey':
                self._data += encode_pubkey(args.pop(0))
            elif item == 'sig':
                r, s = args.pop(0)
                r_str = _deflate_long(r, False)
                s_str = _deflate_long(s, False)
                while len(r_str) < 20:
                    r_str = '\000' + r_str
                while len(s_str) < 20:
                    s_str = '\000' + s_str
                self._data += r_str + s_str
            elif item == 'mac':
                self._data += args.pop(0)[:20]
            elif item == 'tlv':
                t, val = args.pop(0)
                self._data += struct.pack('>HH', t, len(val))
                self._data += val
        return
    
    def unpack(self, format):
        out = []
        n = self._n
        for item in format.split(' '):
            if item == 'byte':
                out.append(struct.unpack('>B', self._data[n:n + 1])[0])
                n += 1
            elif item == 'short':
                out.append(struct.unpack('>H', self._data[n:n + 2])[0])
                n += 2
            elif item == 'int':
                out.append(struct.unpack('>I', self._data[n:n + 4])[0])
                n += 4
            elif item == 'ctr':
                out.append(struct.unpack('>Q', self._data[n:n + 8])[0])
                n += 8
            elif item == 'mpi':
                len = struct.unpack('>i', self._data[n:n + 4])[0]
                n += 4
                out.append(_inflate_long(self._data[n:n + len], True))
                n += len
            elif item == 'blob':
                len = struct.unpack('>i', self._data[n:n + 4])[0]
                n += 4
                out.append(self._data[n:n + len])
                n += len
            elif item == 'pubkey':
                key, offset = decode_pubkey(self._data[n:])
                out.append(key)
                n += offset
            elif item == 'sig':
                r = _inflate_long(self._data[n:n + 20], True)
                n += 20
                s = _inflate_long(self._data[n:n + 20], True)
                n += 20
                out.append((r, s))
            elif item == 'mac':
                out.append(self._data[n:n + 20])
                n += 20
            elif item == 'tlv':
                t, l = struct.unpack('>HH', self._data[n:n + 4])
                val = self._data[n + 4:n + l + 4]
                out.append((t, val))
                n += l + 4
        self._n = n
        return tuple(out)

    @classmethod
    def make(cls, format, *args):
        packet = cls()
        packet.add(format, *args)
        return str(packet)

    @classmethod
    def parse(cls, format, data):
        packet = cls(data)
        return packet.unpack(format)
